/*

Copyright 2001 
by Constantin Loizides <constantin.loizides@isg.de>


This file is part of the journaling-filesystem fragmentation project.
You find the files belonging to that project in the Makefile
coming together with this file.

Source code concerning that project is free software; 
you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  
USA

*/

/* INSTALL
	just untargz and type make agesystem
	agesystem -h gives you options
	
	see also the skript agesystem.sh
*/

#include <gsl/gsl_math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>


#include "common.h"
#include "cl_time.h"


#define PROGNAME "agesystem"
#define VERSION "0.3"

#ifdef EXTRADEFNOTUSED
 #define GAUGEONLY
#endif

/* global vars */
gsl_rng *r;  /* global generator */

/* for common.h */
int verbose=0;
int blocksize=BUFFERSIZE;
char *dummybuffer;


#define N_diff 5
static const int n_timer=(int)(100/N_diff);
cl_Timer timer(n_timer);
/*for mount point info*/
int fdsave=-1; 
cl_String tmpdir="";

void usage(){
	printf("Usage of %s v%s\n",PROGNAME,VERSION);
	printf("\t -b blksize\t\t\tblocksize to use for buffered i/o\n");
	printf("\t -c dir_base:dir_counter\t\t\tstructure of subdirectories to create and use for aging\n");
	printf("\t -d dirname\t\t\tchange to directory \"dirname\"\n");
	printf("\t -f \t\t\t\tuse fixed length of bytes, given in blocksize\n");
	printf("\t -l minimum_file_size_in_bytes\tset minimum file size to minimum_file_size_in_bytes\"\"\n");
	printf("\t -n number_of_files\t\tmaximum number of files to create\n");
	printf("\t -m mean_file_size_in_bytes\tset mean file size to minimum_file_size_in_bytes\"\"\n");
	printf("\t -o access_mode\t\t\tset access mode to  \"access_mode\" \n\t    (eg. cw=%d, cws=%d, cwa=%d,cwas=%d, cwt=%d,cwts=%d)\n",O_CREAT|O_WRONLY,O_CREAT|O_WRONLY|O_SYNC,O_CREAT|O_WRONLY|O_APPEND,O_CREAT|O_WRONLY|O_APPEND|O_SYNC,O_CREAT|O_WRONLY|O_TRUNC,O_CREAT|O_WRONLY|O_TRUNC|O_SYNC);
	printf("\t -h,-?\t\t\t\tthis text\n");
	printf("\t -s file_no\t\t\tstart with filenumber \"file_no\"\n");
	printf("\t -t\t\t\t\ttest input and exit\n");
	printf("\t -u max_fs_usage\t\tmax filesystem usage to reach\n");
	printf("\t -v\t\t\t\tbe verbose\n");
	printf("\t -y seed\t\t\tset random generator seed\n");
	printf("\t -z\t\t\t\tsync with unmount (don't use or check/edit /root/bin/agesync.sh)\n");
	exit(1);
}

float fs_usage(int fdsave){
	
	struct statfs buf_statfs;
	if(fstatfs(fdsave,&buf_statfs)==-1)
	{
		fprintf(stderr,"Error statfsing \"%s\", reason: %s\n",".",strerror(errno));
		exit(1);
	}
	return (float)(100-100.*(float)buf_statfs.f_bfree/(float)buf_statfs.f_blocks);
}


int get_aktion(float fsu,float what_is_100_percent=100)
{	
	/*	1=delete
		2=create
	*/
	if((what_is_100_percent>0)&&(what_is_100_percent<100)){
		fsu=fsu/what_is_100_percent*100;
	}
	
	if(fsu>=98) return 1; /*delete for sure*/
	/*create number between 0 and 100 */
	int rannum=gsl_rng_uniform_int(r,101);
	
	if(fsu<=30){
		if(rannum<=90) return 2; 
		else return 1;
	} else if(fsu<=50){
		if(rannum<=85) return 2; 
		else return 1;
	} else if(fsu<=85){
		if(rannum<=80) return 2;
		else return 1;
	} else if(fsu<=92){
		if(rannum<=50) return 2;
		else return 1;
	} else if(fsu<=98)
	{
		if(rannum<=30) return 2;
		else return 1;
	}	
	
	return -1; /*error*/
}

inline int f2d(int fn,int dircounter)
{
	return (fn % dircounter);
}

inline int d2d(int dn,int dirbase)
{
	return (dn % dirbase);
}


void clsync(int flag=0)
{
	sync();

	if(flag==0) return;
	if(tmpdir.length()==0) return;

	if(fdsave!=-1) close(fdsave);

	if(chdir("/root/bin")==0){
		cl_String cmd="agesync.sh ";
		cmd+=tmpdir;
		//cout << cmd << " ---- " << endl; 
		int ret=system((char *)cmd);
		if(ret!=0){
			fprintf(stderr,"Error call agesync.sh\n");
			exit(1);
		}		
		chdir((char*)tmpdir);
	}

	fdsave=open(".",O_RDONLY);
	if(fdsave==-1){
		fprintf(stderr,"Error opening \"%s\", reason: %s\n",".",strerror(errno));
		exit(1);
	}		
}

int main(int argc,char **argv)
{
	int accessoptions=O_CREAT|O_WRONLY;
	int dircounter=5000;
	int dirbase=100;
	int flag=0;
//	int min_fsu=0;
	int max_fsu=20;
	int min_file_size=0;
	int mean_file_size=1024;
	int min_mfile_size=1024;
	int mean_mfile_size=16384;
	int min_lfile_size=16384;
	int mean_lfile_size=131072;
	int testonly=0;
	int fixedlength=0;	
	long long max_fileno=-1;
	long long start_fileno=-1;
	long seed=0;
		
	/*parse options*/
	int temp;
	char ch;
	while((ch=getopt(argc,argv,"b:c:d:fl:n:m:o:u:h?s:vty:z"))!=EOF){
		switch(ch)	{	
			case 'b':
				temp=atoi(optarg);
				if(temp<1)
				{
					fprintf(stderr, "Can't set blocksize to \"%s\".\n", optarg);
					usage();
				} else blocksize=temp;
				break;
			case 'c':{
				int i=0;
				char *opttok=strtok(optarg,":");
				while((opttok!=NULL)&&(i++<2)){
					//printf("%d-%s\n",++i,opttok);
					temp=atoi(opttok);
					if(temp<0)
					{
						fprintf(stderr, "Can't set directory counter to \"%s\".\n", optarg);
						usage();
					} 
					else if(i==1) dirbase=temp;
					else dircounter=temp;
					opttok=strtok(NULL,":");
				}
				break;}
			case 'd':
				if(chdir(optarg))
				{
					fprintf(stderr, "Can't change to directory \"%s\".\n", optarg);
					usage();
				}
				else tmpdir=optarg;
				break;
			case 'f': fixedlength=1; break;
			case 'm':{
				int i=0;
				char *opttok=strtok(optarg,":");
				while((opttok!=NULL)&&(i++<3)){
					//printf("%d-%s\n",++i,opttok);
					temp=atoi(opttok);
					if(temp<0)
					{
						fprintf(stderr, "Can't set mean file length to \"%s\".\n", optarg);
						usage();
					} else if(i==1) mean_file_size=temp;
					else if(i==2) mean_mfile_size=temp;
					else mean_lfile_size=temp;
					opttok=strtok(NULL,":");
				}
				break;}
			case 'n':
				temp=atol(optarg);
				if(temp<0)
				{
					fprintf(stderr, "Can't set maximum filenumber to \"%s\".\n", optarg);
					usage();
				} else max_fileno=temp;
				break;
			case 'l':{
				int i=0;
				char *opttok=strtok(optarg,":");
				while((opttok!=NULL)&&(i++<3)){
					//printf("%d-%s\n",i,opttok);
					temp=atoi(opttok);
					if(temp<0)
					{
						fprintf(stderr, "Can't set minimum file length to \"%s\".\n", optarg);
						usage();
					} else if(i==1) min_file_size=temp;
					else if(i==2) min_mfile_size=temp;
					else min_lfile_size=temp;
					opttok=strtok(NULL,":");
				}
				break;}
			case 'o':
				temp=atoi(optarg);
				if(temp<1)
				{
					fprintf(stderr, "Can't set access options to \"%s\".\n", optarg);
					usage();
				} else accessoptions=temp;
				break;
			case 'u':
				temp=atoi(optarg);
				if((temp<0)||(temp>100))
				{
					fprintf(stderr, "Can't set max usage to \"%s\".\n", optarg);
					usage();
				} else max_fsu=temp;
				break;
			case 's':
				temp=atol(optarg);
				if(temp<0)
				{
					fprintf(stderr, "Can't set starting filenumber to \"%s\".\n", optarg);
					usage();
				} else start_fileno=temp;
				break;
			case '?':
			case 'h': usage(); break;
			case 'v': verbose=1; break;
			case 't': testonly=1; break;
			case 'y':
				seed=atol(optarg);
				break;
			case 'z': flag=1; break;
			default: usage(); break;
		}
	}

	dummybuffer=new char[blocksize];
	memset(dummybuffer,31,blocksize);

	if(testonly==1){
		cout << "accessoptions: " << accessoptions << endl;
		cout << "dircounter: " << dircounter << endl;
		cout << "max_filenumber: " << max_fileno << endl;
		cout << "start_filenumber: " << start_fileno << endl;
		cout << "dirbase: " << dirbase << endl;
		cout << "min_file_size: " << min_file_size << endl;
		cout << "mean_file_size: " << mean_file_size << endl;
		cout << "min_mfile_size: " << min_mfile_size << endl;
		cout << "mean_mfile_size: " << mean_mfile_size << endl;
		cout << "min_lfile_size: " << min_lfile_size << endl;
		cout << "mean_lfile_size: " << mean_lfile_size << endl;
		cout << "fixedlength: " << fixedlength << endl;
//		cout << "min_fsu: " << min_fsu << endl;
		cout << "max_fsu: " << max_fsu << endl;
		cout << "blocksize: " << blocksize << endl;
		cout << "verbose: " << verbose << endl;
		cout << "flag: " << flag << endl;
		cout << "seed: " << seed << endl;
#ifdef GAUGEONLY
		cout << "GAUGEONLY" << endl;
#endif
		exit(0);
	}

//	gsl_rng_env_setup();
	gsl_rng_default_seed=seed;
	r=gsl_rng_alloc(gsl_rng_default);
		
	fdsave=open(".",O_RDONLY);
	if(fdsave==-1){
		fprintf(stderr,"Error opening \"%s\", reason: %s\n",".",strerror(errno));
		exit(1);
	}		
	

	unsigned long long filecounter=0;
	unsigned long long countedbytes=0;

	//create dirs
	cl_String dir(250);
	for(int i=0;i<dircounter;i++)
	{
		dir="verzeichnis-";
		dir+=d2d(i,dirbase);
		mkdir((char*)dir,S_IRUSR|S_IWUSR|S_IXUSR);
		dir+="/inner-";
		dir+=i;
		int ret=mkdir((char*)dir,S_IRUSR|S_IWUSR|S_IXUSR);
		if(ret==-1){
			if(start_fileno!=-1){
				filecounter=start_fileno;
				break;
			}
			dir="verzeichnis-";
			i=0;
			dir+=d2d(i,dirbase);
			dir+="/inner-";
			dir+=i;
			long no=countdirentries((const char*)dir);
			if(no<0) {
				fprintf(stderr,"Error creating dirs, exiting...\n");
				exit(1);
			} 
			else filecounter=(1+no)*dircounter;
			//printf("Found %llu\n",filecounter);
			break;
		}
	}
	clsync(flag);	/*commit changes*/

	
	float fsu=0.0;
//	int min_fsu=(int)fs_usage(fdsave);
	int which_timer=0;
	int length=0;
	int aktion=0;
	cl_String file(250);		
			
	timer.initialize();
	timer.timestamp();
	while((fsu=fs_usage(fdsave))<max_fsu)
	{
		if(((int)fsu / (int)N_diff)!=which_timer){
			clsync(flag);
			timer.get_delta_t(which_timer,countedbytes);					
			timer.timestamp();
			which_timer=(int)fsu / (int)N_diff;
			countedbytes=0;
		}
		if(verbose==1) printf("Usage: %f with timer %d --- ",fsu,which_timer);

#ifdef cycle
		if((fsu/max_fsu>90)&&(rev==0))rev=1;
		else if((fsu/max_fsu<10)&&(rev==1)) rev=0;

		if(rev==1){ /*invert aktion*/
			aktion=get_aktion(100-fsu,max_fsu);	
			if(aktion==2) aktion=1;
			else aktion=2;	
		} else aktion=get_aktion(fsu,max_fsu);
#else
#ifdef GAUGEONLY
		aktion=2;	/*create only*/
#else		
		aktion=get_aktion(fsu);
#endif
#endif		

		/*delete aktion*/
		if(aktion==1)
		{ 
			int tempcounter=0;
			#define TEMPCOUNTER_ 25
			while(tempcounter<TEMPCOUNTER_)
			{
				long filetodel=0;
				filetodel=gsl_rng_uniform_int(r,filecounter);
				dir="verzeichnis-";
				int dirnumber=f2d(filetodel,dircounter);
				dir+=d2d(dirnumber,dirbase);
				dir+="/inner-";
				dir+=dirnumber;
				file=dir;
				file+="/datei-";
				file+=filetodel;
				int temp=unlinkfile((char*)file);
				if(temp==0) break;
				tempcounter++;				
			}
			if(verbose==1) 
				if(tempcounter<TEMPCOUNTER_) printf("Delete file \"%s\"\n",(char *)file);
				else printf("Delete file missed %d times\n",tempcounter);
		} 
		/*create aktion*/
		else if(aktion==2)
		{ 
			if(max_fileno==-1) filecounter++;
			else filecounter=(filecounter+1) % max_fileno;
			
			int dirnumber=f2d(filecounter,dircounter);
			dir="verzeichnis-";
			dir+=d2d(dirnumber,dirbase);
			dir+="/inner-";
			dir+=dirnumber;
			file=dir;
			file+="/datei-";
			file+=filecounter;

			if(fixedlength==1) length=blocksize;
			else{
				int ran_size=gsl_rng_uniform_int(r,101);
				if(ran_size<90) length=(int)gsl_ran_gaussian_tail(r,min_file_size,mean_file_size); /*bytes*/
				else if(ran_size<=99) length=(int)gsl_ran_gaussian_tail(r,min_mfile_size,mean_mfile_size); /*bytes*/
				else length=(int)gsl_ran_gaussian_tail(r,min_lfile_size,mean_lfile_size); /*bytes*/
			}
			writefile((char *)file,length,accessoptions);
			countedbytes+=length;
			if(verbose) printf("Create file \"%s\" with length %d\n",(char *)file,length);
		} else {
			fprintf(stderr, "Got unknown action: %d\n",aktion);
			exit(1);
		}
	}	

	clsync(flag);
	timer.get_delta_t(which_timer,countedbytes);					
	printf("\n\nTiming Results\n");
	for(int i=0;i<n_timer;i++) timer.print_stat(i);

/*
	for(int i=0;i<dircounter;i++)
	{
		dir="verzeichnis-";
		dir+=i;
		removedir((char *)dir);
	}
*/

	if(fdsave!=-1) close(fdsave);
	delete[] dummybuffer;
	gsl_rng_free(r);
}
