/*

Copyright 2001 
by Constantin Loizides <constantin.loizides@isg.de>


This file is part of the journaling-filesystem fragmentation project.
You find the files belonging to that project in the Makefile
coming together with this file.

Source code concerning that project is free software; 
you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  
USA

*/


#include "cl_time.h"

void cl_Timer::timestamp()
{
	struct rusage res_usage;
	m_last_timestamp=get_cur_time();

	getrusage(RUSAGE_SELF, &res_usage);

/*
	m_last_swaps_stamp=res_usage.ru_nswap;
	m_last_inblocks_stamp=res_usage.ru_inblock;
	m_last_outblock_stamp=res_usage.ru_oublock;

	printf("%lu\n",res_usage.ru_oublock);
	printf("%lu\n",res_usage.ru_nsignals);
	printf("%lu\n",res_usage.ru_maxrss);
	printf("%lu\n",res_usage.ru_msgsnd);	
	printf("%lu\n",res_usage.ru_nvcsw);
	printf("%lu\n",res_usage.ru_nivcsw);
*/
	m_last_cpustamp=TIMEVAL_TO_DOUBLE(res_usage.ru_utime)+TIMEVAL_TO_DOUBLE(res_usage.ru_stime);
}

double cl_Timer::time_so_far()
{
	return get_cur_time()-m_last_timestamp;
}

double cl_Timer::cpu_so_far()
{
	return get_cpu_use()-m_last_cpustamp;
}

double cl_Timer::get_cur_time()
{
	struct timeval tp;
 
	if (gettimeofday(&tp, static_cast<struct timezone *>(NULL)) == -1){
		fprintf(stderr,"Error gettimeofday, reason: %s\n",strerror(errno));
		exit(1);
	}
	return TIMEVAL_TO_DOUBLE(tp);
}

double cl_Timer::get_cpu_use()
{
	struct rusage res_usage;

	getrusage(RUSAGE_SELF, &res_usage);
	return TIMEVAL_TO_DOUBLE(res_usage.ru_utime)+TIMEVAL_TO_DOUBLE(res_usage.ru_stime);
}

void cl_Timer::get_delta_t(int test=0,unsigned long long add=0)
{
	CHECK_BOUNDS(test)
	m_delta[test].Elapsed+=time_so_far();

	struct rusage res_usage;
	getrusage(RUSAGE_SELF, &res_usage);
	
	m_delta[test].CPU+=TIMEVAL_TO_DOUBLE(res_usage.ru_utime)+TIMEVAL_TO_DOUBLE(res_usage.ru_stime)-m_last_cpustamp;
/*
	m_delta[test].swaps+=res_usage.ru_nswap-m_last_swaps_stamp;
	m_delta[test].inblocks+=res_usage.ru_inblock-m_last_inblocks_stamp;
	m_delta[test].outblocks+=res_usage.ru_oublock-m_last_outblock_stamp;
*/
	m_delta[test].counter+=add;
	timestamp();
}

#ifdef notnow
void cl_Timer::get_delta_report(report_s &rep)
{
	rep.EndTime=get_cur_time();
	rep.CPU=cpu_so_far();
	timestamp();
}

void cl_Timer::add_delta_report(report_s &rep,int test)
{
	CHECK_BOUNDS(test)

	if(m_delta[test].CPU==0.0)
	{
		m_delta[test].FirstStart=rep.StartTime;
		m_delta[test].LastStop=rep.EndTime;
	}
	else
	{
		m_delta[test].FirstStart=min(m_delta[test].FirstStart,rep.StartTime);
		m_delta[test].LastStop= max(m_delta[test].LastStop,rep.EndTime);
	}
	m_delta[test].CPU+=rep.CPU;
	m_delta[test].Elapsed=m_delta[test].LastStop-m_delta[test].FirstStart;
}
#endif

cl_Timer::cl_Timer(int T_=1):m_type(txt)
{
	TestCount=T_;
	m_delta=new delta_s[TestCount];
	initialize();
}

cl_Timer::~cl_Timer()
{
	delete[] m_delta;
}

void cl_Timer::initialize()
{
	for(int i=0;i<TestCount;i++)
	{
		m_delta[i].CPU=0.0;
		m_delta[i].Elapsed=0.0;
		m_delta[i].counter=0;
/*
		m_delta[i].swaps=0;
		m_delta[i].inblocks=0;
		m_delta[i].outblocks=0;
*/
	}
	timestamp();
}

void cl_Timer::print_cpu_stat(int test=0,FILE *m_fp=stdout)
{
	CHECK_BOUNDS(test);

	if(m_delta[test].Elapsed==0.0){
		if(m_type==txt)fprintf(m_fp,"    ");
		else fprintf(m_fp,",");
		return;
	}

	if(m_delta[test].Elapsed<MinTime){
		if(m_type==txt)fprintf(m_fp," +++");
		else fprintf(m_fp,",+++");
		return;
	}
  
	int cpu=int(m_delta[test].CPU/m_delta[test].Elapsed*100.0);
	if(m_type==txt) fprintf(m_fp," %3d",cpu);
	else fprintf(m_fp,",%d",cpu);
	return;
}

void cl_Timer::print_time_stat(int test=0,FILE *m_fp=stdout)
{
	if(m_delta[test].Elapsed==0.0){
		if(m_type==txt)fprintf(m_fp, "      ");
		else fprintf(m_fp, ",");
	} else if(m_delta[test].Elapsed<MinTime){
		if(m_type==txt)fprintf(m_fp, " +++++");
    	else fprintf(m_fp, ",+++++");
  	} else {
      double stat=m_delta[test].Elapsed;
      if(m_type == txt) fprintf(m_fp, " %5d", int(stat));
      else fprintf(m_fp, ",%d", int(stat));
	}
}

void cl_Timer::print_stat(int test=0,FILE *m_fp=stdout)
{
	CHECK_BOUNDS(test);

	fprintf(m_fp,"%2d: %e %e %llu\n",test,m_delta[test].Elapsed,m_delta[test].CPU,m_delta[test].counter);
	//fprintf(m_fp,"%2d: %e %e %llu - %llu %llu %llu\n",test,m_delta[test].Elapsed,m_delta[test].CPU,m_delta[test].counter,m_delta[test].swaps,m_delta[test].inblocks,m_delta[test].outblocks);
}


